// ----------------------------------------------------------------------------
// BreakInfo.java
// Line break information for pretty-printing abstract syntax trees.
// $Id: BreakInfo.java,v 1.8 2006/04/05 10:02:50 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Communication;

import fmrisc.Syntax.*;
import fmrisc.Syntax.Number;

import java.util.*;

/**
 * 
 *  Line break information for pretty-printing abstract syntax trees.
 *
 */
public final class BreakInfo
{
  private AST tree;           // the abstract syntax tree 
  private int length;         // the length of its linear representation
  private Vector children;    // infos for the scalar children of the tree
                              // (BreakInfo objects or BreakInfo[] arrays)
 
  // auxiliary variable used during construction
  // true iff mathml representation is considered in computing the length
  private static boolean mathml;
  
  /****************************************************************************
   * Construct the line break information for the denoted tree.
   * @param tree an abstract syntax tree.
   * @param mathml true iff the mathml representation is considered.
   ***************************************************************************/
  public static BreakInfo compute(AST tree, boolean mathml)
  {
    BreakInfo.mathml = mathml;
    return new BreakInfo(tree);
  }
  
  /****************************************************************************
   * Construct the line break information for the denoted tree.
   * @param tree an abstract syntax tree.
   ***************************************************************************/
  private BreakInfo(AST tree)
  {
    this.tree = tree;
    children = new Vector();
    compute(tree);
    if (children.isEmpty()) children = null;
  }
  
  /****************************************************************************
   * Get the tree to which this break information belongs.
   * @return the abstract syntax tree
   ***************************************************************************/
  public AST getTree()
  {
    return tree;
  }
  
  /****************************************************************************
   * Get the total length of the linear representation of the tree.
   * @return the length
   ***************************************************************************/
  public int getLength()
  {
    return length;
  }
  
  /****************************************************************************
   * Get the children of this node.
   * @return a vector holding BreakInfo objects and BreakInfo[] arrays.
   ***************************************************************************/
  public Vector getChildren()
  {
    return children;
  }
  
  /***************************************************************************
   * Get sum of the lengths of the info nodes
   **************************************************************************/
  public static int length(BreakInfo[] infos)
  {
    int s = 0;
    int n = infos.length;
    for (int i=0; i<n; i++)
      s = s+infos[i].getLength();
    return s;
  }
  
  /***************************************************************************
   * Add to break information a new child
   * @param child an abstract syntax tree who is the child of this tree
   *        (may be null, then the child is ignored).
   **************************************************************************/
  private void addChild(AST child)
  {
    if (child == null) return;
    BreakInfo info = new BreakInfo(child);
    children.add(info);
  }
  
  /****************************************************************************
   * Add to break information an array of children.
   * @param achild an array of abstract syntax trees which is the child
   *        of this tree (may be null, then the array is ignored).
   ***************************************************************************/
  private void addChildren(AST[] achild)
  {
    if (achild == null) return;
    int n = achild.length;
    BreakInfo[] infos = new BreakInfo[n];
    for (int i=0; i<n; i++)
      infos[i] = new BreakInfo(achild[i]);
    children.add(infos);
  }
  
  /***************************************************************************
   * Set this length to the length of denoted text plus the children lengths.
   * @param text the text added by this abstract syntax tree.
   ***************************************************************************/
  private void setLength(String text)
  {
    length = text.length() + getChildrenLengths();
  }
  
  /****************************************************************************
   * Get the total length of the linear representations of the children.
   * @return sum of the lengths of the children representations.
   ***************************************************************************/
  private int getChildrenLengths()
  {
    if (children == null) return 0;
    int sum = 0;
    int n = children.size();
    for (int i=0; i<n; i++)
    {
      Object elem = children.elementAt(i);
      if (elem instanceof BreakInfo)
      {
        sum = sum + ((BreakInfo)elem).getLength();
        continue;
      }
      BreakInfo[] info = (BreakInfo[])elem;
      if (info == null) continue;
      int m = info.length;
      for (int j=0; j<m; j++)
        sum = sum + info[j].getLength();
    }
    return sum;
  }
  
  /****************************************************************************
   * Adjust length by the separators between an array of trees.
   * @param trees the array of trees separated.
   * @param separator the separation string.
   ***************************************************************************/
  private void adjustArray(AST[] trees, String separator)
  { 
    int n = trees.length;
    if (n > 1) length = length + separator.length()*(n-1);
  }
  
  /***************************************************************************
   * Adjust length by the parentheses encapsulating an expression
   * @param exp an expression
   * @param treshold the priority of the expression in which exp occurs.
   *************************************************************************/
  private void adjustPriority(Expression exp, int treshold)
  {
    int p = exp.getPriority();
    if (p > treshold) length = length + 2; // "()"
  }
  
  /****************************************************************************
   * Compute this break information for denoted tree.
   * @param tree an abstract syntax tree
   ***************************************************************************/
  private void compute(AST tree)
  {
    if (tree instanceof Identifier) 
      computeIdentifier((Identifier)tree);
    else if (tree instanceof Number)
      computeNumber((Number)tree);
    else if (tree instanceof Expression)
      computeExpression((Expression)tree);
    else if (tree instanceof Type)
      computeType((Type)tree);
    else if (tree instanceof Declaration)
      computeDeclaration((Declaration)tree);
    else if (tree instanceof Selector)
      computeSelector((Selector)tree);
    else if (tree instanceof TypedIdentifier)
      computeTypedIdentifier((TypedIdentifier)tree);
    else if (tree instanceof ValuedIdentifier)
      computeValuedIdentifier((ValuedIdentifier)tree);
    else
    {
      System.out.println("Unknown syntax tree to break: " + tree.toString());
      System.exit(-1);
    }
  }
  
  private void computeIdentifier(Identifier ident)
  {
    setLength(ident.getName());
  }
  
  private void computeNumber(Number number)
  {
    setLength(number.getDigits());
  }
  
  private void computeType(Type type)
  {
    if (type instanceof AtomicType)
    {
      AtomicType t = (AtomicType)type;
      setLength(t.getName());
      return;
    }
    if (type instanceof FunctionType)
    {
      FunctionType t = (FunctionType)type;
      Type[] domain = t.getDomain();
      int n = domain.length;
      addChildren(domain);
      addChild(t.getRange());
      if (n == 1)
      {
        if (domain[0] instanceof FunctionType ||
            domain[0] instanceof ArrayType)
          setLength("()->");
        else
          setLength("->");
      }
      else
      {
        setLength("()->");
        adjustArray(domain, ", ");
      }
      return;
    }
    if (type instanceof TupleType)
    {
      TupleType t = (TupleType)type;
      Type[] base = t.getBase();
      addChildren(base);
      setLength("[]");
      adjustArray(base, ", ");
      return;
    }
    if (type instanceof RecordType)
    {
      RecordType t = (RecordType)type;
      TypedIdentifier[] base = t.getBase();
      addChildren(base);
      setLength("[##]");
      adjustArray(base, ", ");
      return;
    }
    if (type instanceof ArrayType)
    {
      ArrayType t = (ArrayType)type;
      addChild(t.getIndex());
      Type base = t.getBase();
      addChild(base);
      setLength("ARRAY  OF ");
      return;
    }
    if (type instanceof SubrangeType)
    {
      SubrangeType t = (SubrangeType)type;
      addChild(t.getFrom());
      addChild(t.getTo());
      setLength("[..]");
      return;
    }
    if (type instanceof SubType)
    {
      SubType t = (SubType)type;
      addChild(t.getPredicate());
      setLength("SUBTYPE()");
      return;
    }
    if (type instanceof LetType)
    {
      LetType t = (LetType)type;
      Declaration[] d = t.getDeclarations();
      addChildren(d);
      addChild(t.getBase());
      setLength("LET  IN ");
      adjustArray(d, ", ");
      return;
    }
    System.out.println("Unknown type to break: " + type.toString());
    System.exit(-1);
  }
  
  private void computeDeclaration(Declaration decl)
  {
    Identifier ident = decl.getIdentifier();
    addChild(ident);
    if (decl instanceof TypeDeclaration)
    {
      TypeDeclaration d = (TypeDeclaration)decl;
      Type t = d.getType();
      if (t == null)
        setLength(": TYPE");
      else
      {
        addChild(t);
        setLength(": TYPE = ");
      }
      return;
    }
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration d = (ValueDeclaration)decl;
      Type t = d.getType();
      Expression e = d.getValue();
      if (t == null)
      {
        addChild(e);
        setLength(" = ");
      }
      else if (e == null)
      {
        addChild(t); 
        setLength(":");
      }
      else
      {
        addChild(t);
        addChild(e);
        setLength(": = ");
      }
      return;
    }
    if (decl instanceof FormulaDeclaration)
    {
      FormulaDeclaration d = (FormulaDeclaration)decl;
      addChild(d.getFormula());
      if (d instanceof AxiomDeclaration)
        setLength(": AXIOM ");
      else
        setLength(": FORMULA ");
      return;
    }
    System.out.println("Unknown declaration to break: " + decl.toString());
    System.exit(-1);
  }
  
  private void computeSelector(Selector sel)
  { 
    if (sel instanceof SelectorIdentifier)
    {
      SelectorIdentifier s = (SelectorIdentifier)sel;
      addChild(s.getIdentifier());
      setLength(".");
      return;
    }
    if (sel instanceof SelectorIndex)
    {
      SelectorIndex s = (SelectorIndex)sel;
      addChild(s.getIndex());
      setLength("[]");
      return;
    }
    if (sel instanceof SelectorNumber)
    {
      SelectorNumber s = (SelectorNumber)sel;
      addChild(s.getNumber());
      setLength(".");
      return;
    }
    System.out.println("Unknown selector to break: " + sel.toString());
    System.exit(-1);
  }
  
  private void computeTypedIdentifier(TypedIdentifier tident)
  {
    addChild(tident.getIdentifier());
    addChild(tident.getType());
    if (mathml) 
      setLength(": "); // extra space for epsilon sign
    else 
      setLength(":");
  }
  
  private void computeValuedIdentifier(ValuedIdentifier vident)
  {
    addChild(vident.getIdentifier());
    addChild(vident.getValue());
    setLength(":=");
  }

  private void computeExpression(Expression exp)
  {
    if (exp instanceof ApplicationExpression)
      computeApplicationExpression((ApplicationExpression)exp);
    else if (exp instanceof Logical)
      computeLogical((Logical)exp);
    else if (exp instanceof BinaryExpression)
      computeBinaryExpression((BinaryExpression)exp);
    else if (exp instanceof UnaryExpression)
      computeUnaryExpression((UnaryExpression)exp);
    else if (exp instanceof IfThenElseExpression)
      computeIfThenElseExpression((IfThenElseExpression)exp);
    else if (exp instanceof LetExpression)
      computeLetExpression((LetExpression)exp);
    else if (exp instanceof QuantifiedExpression)
      computeQuantifiedExpression((QuantifiedExpression)exp);
    else if (exp instanceof RecordTerm)
      computeRecordTerm((RecordTerm)exp);
    else if (exp instanceof TupleTerm)
      computeTupleTerm((TupleTerm)exp);
    else if (exp instanceof SelectionTerm)
      computeSelectionTerm((SelectionTerm)exp);
    else if (exp instanceof UpdateTerm)
      computeUpdateTerm((UpdateTerm)exp);
    else
    {
      System.out.println("Unknown expression to break: " + exp.toString());
      System.exit(-1);
    }
  }
  
  private void computeApplicationExpression(ApplicationExpression exp)
  {
    addChild(exp.getFunction());
    Expression[] args = exp.getArguments();
    addChildren(args);
    setLength("()");
    adjustArray(args, ", ");
  }
  
  private void computeLogical(Logical logical)
  {
    if (logical.getValue())
      setLength("TRUE");
    else
      setLength("FALSE");
  }
  
  private void computeBinaryExpression(BinaryExpression exp)
  {
    Expression first = exp.getFirst();
    Expression second = exp.getSecond();
    addChild(first);
    addChild(second);
    int p = exp.getPriority();
    if (exp instanceof EquivalentFormula || exp instanceof NotEquivalentFormula)
    {
      setLength(" <=> ");
      if (first instanceof EquivalentFormula ||
          first instanceof NotEquivalentFormula ||
          first instanceof ImpliesFormula ||
          first.getPriority() > p)
        length = length+2; // "()"
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula ||
          second instanceof ImpliesFormula ||
          second.getPriority() > p)
        length = length+2; // "()"
      return;
    }
    if (exp instanceof ImpliesFormula)
    {
      setLength(" => ");
      if (first instanceof EquivalentFormula ||
          first instanceof NotEquivalentFormula ||
          first instanceof ImpliesFormula ||
          first.getPriority() > p)
        length = length+2; // "()"
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula ||
          second.getPriority() > p)
        length = length+2; // "()"
      return;
    }
    if (exp instanceof PowerTerm)
    {
      setLength("^");
      adjustPriority(first, p);
      if (second instanceof PowerTerm ||
          second.getPriority() > p)
        length = length+2; // "()"
      return;
    }
    if (exp instanceof AndFormula)
    {
      if (mathml)
        setLength(" & ");
      else
        setLength(" AND ");
    }
    else if (exp instanceof DividesTerm)
      setLength("/");
    else if (exp instanceof EqualsFormula)
      setLength(" = ");
    else if (exp instanceof GreaterEqualFormula)
      setLength(" >= ");
    else if (exp instanceof GreaterFormula)
      setLength(" > ");
    else if (exp instanceof LessEqualFormula)
      setLength(" <= ");
    else if (exp instanceof LessFormula)
      setLength(" < ");
    else if (exp instanceof MinusTerm)
      setLength("-");
    else if (exp instanceof NotEqualsFormula)
      setLength(" /= ");
    else if (exp instanceof OrFormula)
    {
      if (mathml)
        setLength(" | ");
      else
        setLength(" OR ");
    }
    else if (exp instanceof PlusTerm)
      setLength("+");
    else if (exp instanceof TimesTerm)
      setLength("*");
    else
    {
      System.out.println("Unknown binary expression to break: " + exp.toString());
      System.exit(-1);
    }
    adjustPriority(first, p);
    adjustPriority(second, p);
  }
  
  private void computeUnaryExpression(UnaryExpression exp)
  {
    Expression base = exp.getBase();
    addChild(base);
    int p = exp.getPriority();
    if (exp instanceof NegationTerm)
      setLength("-");
    else if (exp instanceof NotFormula)
      setLength("NOT ");
    else
    {
      System.out.println("Unknown unary expression to break: " + exp.toString());
      System.exit(-1);
    }
    adjustPriority(base, p);
  }
  
  private void computeIfThenElseExpression(IfThenElseExpression exp)
  {
    addChild(exp.getCondition());
    addChild(exp.getThenBranch());
    addChild(exp.getElseBranch());
    if (exp.getElseBranch() instanceof IfThenElseExpression)
      setLength("IF  THEN  ELS ENDIF");
    else
      setLength("IF  THEN  ELSE  ENDIF");
  }
  
  private void computeQuantifiedExpression(QuantifiedExpression exp)
  {
    TypedIdentifier[] vars = exp.getVariables();
    Expression base = exp.getBase();
    addChildren(vars);
    addChild(base);
    if (exp instanceof ForallFormula)
    {
      if (mathml)
        setLength("A(): ");
      else
        setLength("FORALL(): ");
    }
    else if (exp instanceof ExistsFormula)
    {
      if (mathml)
        setLength("E(): ");
      else
        setLength("EXISTS(): ");
    }
    else if (exp instanceof ArrayTerm)
      setLength("ARRAY(): ");
    else if (exp instanceof LambdaTerm)
    {
      if (mathml)
        setLength("L(): ");
      else
        setLength("LAMBDA(): ");
    }
    else
    {
      System.out.println("Unknown quantified expression to break: " + exp.toString());
      System.exit(-1);
    }
    adjustArray(vars, ", ");
    adjustPriority(base, exp.getPriority());    
  }
  
  private void computeLetExpression(LetExpression exp)
  {
    Declaration[] d = exp.getDeclarations();
    addChildren(d);
    addChild(exp.getBase());
    setLength("LET  IN ");
    adjustArray(d, ", ");
  }
  
  private void computeRecordTerm(RecordTerm exp)
  {
    ValuedIdentifier[] base = exp.getBase();
    addChildren(base);
    setLength("(##)");
    adjustArray(base, ", ");
  }
  
  private void computeTupleTerm(TupleTerm exp)
  {
    Expression[] base = exp.getBase();
    addChildren(base);
    setLength("()");
    adjustArray(base, ", ");
  }
  
  private void computeSelectionTerm(SelectionTerm exp)
  {
    Expression base = exp.getBase();
    addChild(base);
    addChild(exp.getSelector());
    setLength("");
    adjustPriority(base, exp.getPriority());
  }
  
  private void computeUpdateTerm(UpdateTerm exp)
  {
    Expression base = exp.getBase();
    Selector[] sels = exp.getSelectors();
    Expression update = exp.getUpdate();
    addChild(base);
    addChildren(sels);
    addChild(update);
    setLength(" WITH :=");
    int p = exp.getPriority();
    adjustPriority(base, p);
    adjustArray(sels, ", ");
    adjustPriority(update, p);
  }
}
