// ----------------------------------------------------------------------------
// Induction.java
// The "induction" command.
// $Id: Induction.java,v 1.20 2007/03/13 08:58:22 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.PrintWriter;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fmrisc.ProofNavigator.State;
import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

/**
 * 
 * Command "induction": induction on a univerally quantified NAT/INT variable.
 *
 */
public final class Induction extends ProofCommandBase
{
  public final static String Name = "induction";
  private Identifier var; // the induction variable (may be null)
  private String label;   // the formula label (may be null)
  
  /***************************************************************************
   * Create an "induction" command.
   * @param var the variable on which to run the induction or null
   *            (if null, induction runs on the first INTEGER variable).
   * @param label the label of the goal to be proved by induction.
   *            (if null, induction runs on the single goal)
   **************************************************************************/
  public Induction(Identifier var, String label)
  {
    super(Name);
    this.label = label;
    this.var = var;
  }
  
  /****************************************************************************
   * Process command "induction": run induction on a goal univerally
   * quantified by a natural number or integer variable.
   * 
   * e: A, ... |- FORALL(...,x:INT,...): p, ... ->
   *  (e: A, ..., FORALL(...,x:INT,...): x>=0 => p |-
   *     FORALL(...,x:INT,...): p, ...) (not used for x:NAT)
   *   e: A, ... |- FORALL(...): p[0/x], ...
   *   e[a:NAT]: A, ..., FORALL(...): p[a/x] |- FORALL(...): p[a+1/x], ...
   *   
   * @param state the proof state to which the command is applied
   * @return null or array of two or three new proof states.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Environment env = state.getEnvironment();
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    int gn = goals.length;
    
    Formula goal;
    if (label == null)
    {
      if (gn != 1)
      {
        error("there must be a single goal");
        return null;
      }
      goal = goals[0];
    }
    else
    {
      goal = state.getGoal(label);
      if (goal == null)
      {
        error("there is no goal [" + label + "] in current state");
        return null;
      }
    }
    
    Formula[] goals0= new Formula[gn-1];
    int gi = 0;
    for (int i=0; i<gn; i++)
    {
      if (goals[i] == goal) continue;
      goals0[gi] = goals[i];
      gi = gi+1;
    }

    Expression exp = goal.getExpression();
    if (!(exp instanceof ForallFormula))
    {
      error("goal [" + label + "] is not a FORALL formula");
      return null;
    }
    ForallFormula exp0 = (ForallFormula)exp;
    TypedIdentifier[] vars = exp0.getVariables();
    Expression body = exp0.getBase();
    
    // find induction variable
    String name = null;
    if (var != null) name = var.getName();
    int n = vars.length;
    int p = -1;
    for (int i=0; i<n; i++)
    {
      ValueDeclIdentifier ident = vars[i].getIdentifier();
      if (name == null)
      {
        Type type = ident.getSymbol().getType();
        if (Checking.isNat(type) || Checking.isInt(type))
        {
          p=i;
          break;
        }
      }
      else if (ident.getName().equals(name))
      {
        Type type = ident.getSymbol().getType();
        if (!Checking.isNat(type) && !Checking.isInt(type))
        {
          error("Cannot run induction variable " + 
            vars[i].toString() + " (type NAT or INT expected)");
          return null;
        }
        p = i;
        // replace identifier by symbol identifier for storing command
        var = ident;
        break;
      }
    }
    if (p == -1)
    {
      if (name == null)
        error("No variable of type NAT or INT in goal [" + label + "]");
      else
        error("No variable with name " + name + " in goal [" + label + "]");
      return null;
    }
    
    // fix induction variable and body
    ValueDeclIdentifier ivar = vars[p].getIdentifier();
    if (n > 1)
    {
      TypedIdentifier[] vars0 = new TypedIdentifier[n-1];
      for (int i=0; i<p; i++)   vars0[i] = vars[i];
      for (int i=p; i<n-1; i++) vars0[i] = vars[i+1];
      body = Construct.forallFormula(vars0, body);
    }
    
    // if variable is of type INT, build additinal case
    ProofState state2 = null;
    if (Checking.isInt(ivar.getSymbol().getType()))
    {
      state2 = new ProofState(state, env);
      Formula ass0 = new Formula(state2,
        Construct.forallFormula(new TypedIdentifier[] { vars[p] }, 
          Construct.impliesFormula(
            Construct.greaterEqualFormula(Construct.reference(ivar),
              Construct.number("0")), 
          body)), false);
      state2.setAssumptions(state2.add(assumptions, ass0));
      state2.setGoals(state2.add(goals0, new Formula(state2, goal)));
    }
    
    // build base case
    ProofState state0 = new ProofState(state, env);
    state0.setAssumptions(state0.copy(assumptions));
    Expression body0 = ProofUtil.bind(ivar, Construct.number("0"), body);
    Formula base = new Formula(state0, body0, true);
    state0.setGoals(state0.add(goals0, base));

    // build induction step
    Environment env0 = env.cloneGlobal();
    ProofState state1 = new ProofState(state, env0); 
    ValueDeclIdentifier ident0 = ivar.instantiate();
    ident0.makeNameUnique(state.getSkolemTable());
    Type type0 = Checking.getNatType();
    ValueSymbol symbol = ident0.getSymbol();
    symbol.setType(type0);
    Expression value0 = Construct.reference(ident0);
    TypeExpression te = Checking.clearSubtypes(type0, value0);
    symbol.setClearType(te);
    env0.put(ident0);
    Expression body1 = ProofUtil.bind(ivar, value0, body);
    Formula hyp = new Formula(state1, body1, false);
    state1.setAssumptions(state0.add(assumptions, hyp));
    Expression value1 =
      Construct.plusTerm(Construct.reference(ident0), Construct.number("1"));
    Expression body2 = ProofUtil.bind(ivar, value1, body);
    Formula step = new Formula(state1, body2, true);
    state1.setGoals(state1.add(goals0, step));
    if (state2 == null)
      return new ProofState[] { state0, state1 };
    else
      return new ProofState[] { state2, state0, state1 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    if (var == null) return;
    Node node1 = toNode(var);
    node.appendChild(node1);
    if (label == null) return;
    Node node0 = newNode("label");
    node0.appendChild(newTextNode(label));
    node.appendChild(node0);
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int l = children.getLength();
    if (l > 2)
    {
      State.setError("at most two children expected: " + node.toString());
      return null;
    }
    if (l == 0) return new Induction(null, null);
    Identifier var = toIdentifier(children.item(0));
    if (var == null) return null;
    if (l == 1) return new Induction(var, null);
    NodeList children0 = children.item(1).getChildNodes();
    if (children0.getLength() != 1) return null;
    String label = toString(children0.item(0));
    return new Induction(var, label);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (var == null) return;
    out.print(" ");
    var.print(out);
    if (label != null)
    {
      out.print(" in ");
      out.print(label);
    }
  }
}
