// ----------------------------------------------------------------------------
// ProofTree.java
// Graphical representation of proof.
// $Id: ProofTree.java,v 1.8 2005/12/27 12:16:41 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.SWT;

import fmrisc.Proving.*;
import fmrisc.ProofNavigator.Commands.*;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.widgets.*;

/**
 * 
 * Graphical representation of proof.
 *
 */
public final class ProofTree 
{
  private Proof proof; // the proof
  private Tree tree;   // its visual representation

  /****************************************************************************
   * Construct proof tree for denoted parent component and with denoted style..
   * @param parent the component to which the tree is attached.
   ***************************************************************************/
  public ProofTree(Composite parent)
  {
    tree = new Tree(parent, SWT.SINGLE);

    // if item is expanded, expand all subnodes
    tree.addTreeListener(new TreeAdapter() {
      public void treeCollapsed(TreeEvent e) {
      }
      public void treeExpanded(TreeEvent e) {
        final TreeItem item = (TreeItem)e.item;
        item.getDisplay().asyncExec(new Runnable() {
          public void run() {
            expandAll(item);
          }
        });
      }
    });

    tree.addSelectionListener(new SelectionAdapter() {
      // if item is double-clicked, goto denoted proof state
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        ProofState state = searchState(item, proof.getRoot());
        MainSWT.process(new Goto(state.getLabel()));
      }
      
      // if item is selected, display corresponding proof state
      public void widgetSelected(SelectionEvent e) {
        TreeItem item = (TreeItem)e.item;
        ProofState state = searchState(item, proof.getRoot());
        MainSWT.setProofState(state);
      }
    });
  }
  
  /****************************************************************************
   * Expand all tree items below item.
   * @param item the root of the subtree to be expanded.
   ***************************************************************************/
  private static void expandAll(TreeItem item)
  {
    item.setExpanded(true);
    int n = item.getItemCount();
    for (int i=0; i<n; i++)
      expandAll(item.getItem(i));
  }
  
  /****************************************************************************
   * Search state related to item in state tree with denoted root.
   * @param item the tree item.
   * @param root the root of the state tree to be searched
   * @return the state associated to item (null, if none found).
   **************************************************************************/
  private static ProofState searchState(TreeItem item, ProofState root)
  {
    if (root == null) return null;
    ProofTreeItem pitem = (ProofTreeItem)root.getListener();
    if (pitem.getTreeItem() == item) return root;
    ProofState[] children = root.getChildren();
    if (children == null) return null;
    int n = children.length;
    for (int i=0; i<n; i++)
    {
      ProofState result = searchState(item, children[i]);
      if (result != null) return result;
    }
    return null;
  }
      
  /***************************************************************************
   * Change tree to visualize denoted proof.
   * @param proof to be visualized.
   **************************************************************************/
  public void setProof(Proof proof)
  {
    this.proof = proof;
    tree.getDisplay().syncExec(new Runnable() {
      public void run() {
        tree.removeAll();
      }
    });
    if (proof != null)
      ProofTreeItem.setNodes(tree, proof.getRoot());
  }
  
  /***************************************************************************
   * Signal run status of background process.
   * @param running if we are waiting for background process to complete.
   **************************************************************************/
  public void setRunning(final boolean running)
  {
    tree.getDisplay().syncExec(new Runnable() {
      public void run() {
        tree.setEnabled(!running);
      }
    });
  }
  
  /***************************************************************************
   * Set all states in tree widget to unselected.
   **************************************************************************/
  public void setUnselected()
  {
    tree.getDisplay().syncExec(new Runnable() {
      public void run() {
        tree.setSelection(new TreeItem[] { });
      }
    });
  }
}
