// ----------------------------------------------------------------------------
// Proof.java
// A proof represented by a tree of proof states.
// $Id: Proof.java,v 1.44 2009/04/02 10:56:22 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving;

import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.Communication.*;

import java.util.*;
import java.io.*;

import org.w3c.dom.*;

/*
 * A proof represented by a tree of proof states.
 */
public final class Proof
{
  /** 
   * A proof contains the following pieces of status information:
   * 
   * - skeleton:   true iff proof is just skeleton that needs to be replayed.
   * - trusted:    true iff proof is not a skeleton or 
   *               proof is skeleton but does not depend on entities 
   *               (types, values, formulas) whose declarations have changed 
   *               since the proof has been replayed for the last time.
   *               Replaying a trusted proof must not cause errors.
   * - closed:     true iff there are no open states in the proof.
   * - absolute:   true iff the proof is trusted and closed and only depends 
   *               on formulas with absolute proofs.
   */

  private boolean skeleton; // proof is skeleton
  private boolean trusted;  // proof is trusted
  private boolean closed;   // proof is closed
  private boolean absolute; // proof is absolute
  
  private FormulaSymbol symbol;     // the symbol of the formula proved
  private ProofState root;             // the root of the proof state tree
  private HashMap states = new HashMap(); // the collection of all tree states
  private Vector open = new Vector();  // the collection of open leaf states
  private ProofState current;          // the current open proof state
  
  // output stream
  private static PrintWriter out = State.getSystemOut();
  
  /***************************************************************************
   * Construct new proof for formula denoted by symbol.
   * @param symbol a formula symbol
   * @param autoSimplify true iff automatic simplification is switched on
   * @return the proof (null, if something went wrong)
   *************************************************************************/
  public static Proof newProof(FormulaSymbol symbol, boolean autoSimplify)
  {
    Proof proof = new Proof(symbol, autoSimplify);
    boolean okay = proof.initialize();
    if (!okay) return null;
    return proof;
  }
  
  /***************************************************************************
   * Construct new proof for formula denoted by symbol.
   * @param symbol a formula symbol
   * @param autoSimplify true iff automatic simplification is switched on
   *************************************************************************/
  private Proof(FormulaSymbol symbol, boolean autoSimplify)
  {
    this.symbol = symbol;
    Environment env = symbol.getEnvironment();
    root = new ProofState(this, env, symbol.getFormula(), autoSimplify);
  }
  
  /***************************************************************************
   * Initialize the proof state from root state.
   * @return true iff everything went okay.
   **************************************************************************/
  private boolean initialize()
  {
    skeleton = false;
    trusted = true;
    closed = false;
    absolute = true;
    states.put(root.getLabel(), root);
    boolean auto = State.useAutoSimplify();
    State.setAutoSimplify(root.useAutoSimplify());
    boolean okay = root.simplify(false);
    State.setAutoSimplify(auto);
    if (!okay) return false;
    if (root.isClosed())
    {
      out.println("Proof is completed by simplification."); 
      closed = true;
    }
    else
    {
      open.add(root);
      current = root;
    }
    return true;
  }
  
  /***************************************************************************
   * Construct skeleton proof.
   *************************************************************************/
  private Proof()
  {
    skeleton = true;
    trusted = false;
    closed = false;
    absolute = false;
  }
  
  /***************************************************************************
   * Is proof skeleton?
   * @return true iff proof is skeleton.
   *************************************************************************/
  public boolean isSkeleton()
  {
    return skeleton;
  }
  
  /***************************************************************************
   * Signal whether proof is skeleton.
   * @param skeleton true iff proof is skeleton.
   **************************************************************************/
  public void setSkeleton(boolean skeleton)
  {
    this.skeleton = skeleton;
  }
  
  /***************************************************************************
   * Is proof trusted?
   * @return true iff proof is trusted.
   *************************************************************************/
  public boolean isTrusted()
  {
    return trusted;
  }
  
  /***************************************************************************
   * Signal whether proof is trusted.
   * @param trusted true iff proof is trusted.
   **************************************************************************/
  public void setTrusted(boolean trusted)
  {
    this.trusted = trusted;
  }
  
  /***************************************************************************
   * Is proof closed?
   * @return true iff proof is closed.
   *************************************************************************/
  public boolean isClosed()
  {
    return closed;
  }
  
  /***************************************************************************
   * Signal whether proof is closed.
   * @param closed true iff proof is closed.
   **************************************************************************/
  public void setClosed(boolean closed)
  {
    this.closed = closed;
  }
  
  /***************************************************************************
   * Is proof absolute?
   * @return true iff proof is absolute.
   *************************************************************************/
  public boolean isAbsolute()
  {
    return absolute;
  }
  
  /***************************************************************************
   * Signal whether proof is closed.
   * @param absolute true iff proof is closed.
   **************************************************************************/
  public void setAbsolute(boolean absolute)
  {
    this.absolute = absolute;
  }
   
  /***************************************************************************
   * get symbol of formula proved
   * @return the symbol of the formula proved
   **************************************************************************/
  public FormulaSymbol getSymbol()
  {
    return symbol;
  }
  
  /***************************************************************************
   * get root of proof
   * @return proof state that represents the root
   **************************************************************************/
  public ProofState getRoot()
  {
    return root;
  }
  
  /***************************************************************************
   * get current open proof state
   * @return current open proof state (null, if there is none)
   **************************************************************************/
  public ProofState getCurrent()
  {
    return current;
  }
  
  /***************************************************************************
   * set current open proof state
   * @param state an open proof state
   **************************************************************************/
  public void setCurrent(ProofState state)
  {
    ProofState old = current;
    current = state;
    // must first reset current such that display is updated correctly
    if (old != null && (current == null || current.getParent() != old))
      old.hasChanged(false);
    if (current != null) 
      current.hasChanged(false);
  }
  
  /***************************************************************************
   * get open leaf goal
   * @return some open leaf goal (null, if there is none)
   **************************************************************************/
  public ProofState getOpenGoal()
  {
    if (open.isEmpty()) return null;
    return (ProofState)(open.elementAt(0));
  }
  
  /***************************************************************************
   * get open leaf goals
   * @return array of open leaf goals 
   **************************************************************************/
  public ProofState[] getOpenGoals()
  {
    return ProofUtil.toProofStateArray(open);
  }
  
  /***************************************************************************
   * Get proof state with denoted tag.
   * @param tag the tag of the state
   * @return the state with this tag (null, if there is no such state)
   *************************************************************************/
  public ProofState getState(String tag)
  {
    return (ProofState)states.get(tag);  
  }
    
  /****************************************************************************
   * Collect symbols of all formulas on which the proof tree at the 
   * denoted root depends.
   * @param symbols vector to which the formula symbols are added.
   ***************************************************************************/
  private void collectDependencies(ProofState state, Vector symbols)
  {
    Command command = state.getCommand();
    if (command == null) return;
    if (command instanceof Lemma)
    {
      // lookup lemma symbols in global environment
      // (proof state cannot add lemmas)
      FormulaSymbol[] symbols0 = ((Lemma)command).getSymbols(State.getEnvironment());
      int n = symbols0.length;
      for (int i=0; i<n; i++)
        symbols.addElement(symbols0[i]);
    }
    ProofState[] children = state.getChildren();
    if (children == null) return;
    int n = children.length;
    for (int i=0; i<n; i++)
      collectDependencies(children[i], symbols);
  }
  
  /****************************************************************************
   * The status of this proof or of the proof of a formula on which this 
   * proof depends may has changed. Compute and set the absoluteness state.
   ***************************************************************************/
  public void computeAbsolute()
  {
    // positive base assumption
    absolute = true;
    
    // if proof is not closed, it is not absolute
    if (!closed) 
    {
      absolute = false;
      return;
    }
    
    // if proof depends on a non-absolute proof, it is not absolute
    Vector symbols = new Vector();
    collectDependencies(root, symbols);
    int n = symbols.size();
    for (int i=0; i<n; i++)
    {
      FormulaSymbol symbol = (FormulaSymbol)symbols.elementAt(i);
      if (symbol.isAxiom()) continue;
      Proof proof = symbol.getProof();
      if (proof == null || !proof.isAbsolute())
      {
        absolute = false;
        return;
      }
    }
  }
  
  /***************************************************************************
   * Cut proof at denoted state.
   * @param state state at which the proof is cut
   **************************************************************************/
  public void cut(ProofState state)
  {
    setCurrent(state);
    removeChildren(state);
    state.open();
    open = new Vector();
    root.addOpen(open);
    closed = open.isEmpty();
  }
  
  /***************************************************************************
   * Remove children of denoted state (recursively) from list of states.
   * @param state state whose children are to be removed.
   **************************************************************************/
  private void removeChildren(ProofState state)
  {
    ProofState[] children = state.getChildren();
    if (children == null) return;
    int n = children.length;
    for (int i=0; i<n; i++)
    {
      ProofState child = children[i];
      removeChildren(child);
      states.remove(child.getLabel());
    }
  }
  
  /***************************************************************************
   * Inform proof that state was closed.
   * @param state a proof state that was closed
   **************************************************************************/
  public void close(ProofState state)
  {
    open.remove(state);
    if (open.isEmpty())
    {
      closed = true;
      setCurrent(null);
    }
    else
      setCurrent((ProofState)open.firstElement());
  }
  
  /***************************************************************************
   * Add children of denoted state (recursively) to list of (also open) states.
   * @param state state whose children are to be added.
   **************************************************************************/
  private void addChildren(ProofState state)
  {
    ProofState[] children = state.getChildren();
    if (children == null) return;
    int n = children.length;
    for (int i=0; i<n; i++)
    {
      ProofState child = children[i];
      states.put(child.getLabel(), child);
      if (!child.isClosed()) open.add(child);
      addChildren(child);
    }
  }
  
  /***************************************************************************
   * Redo proof at denoted state.
   * @param state state at which the proof is redone.
   * @return true iff proof could be redone
   **************************************************************************/
  public boolean redo(ProofState state)
  {
    state.redo();
    if (state.getCommand() == null) return false;
    addChildren(state);
    open = new Vector();
    root.addOpen(open);
    closed = open.isEmpty();
    ProofState[] children = state.getChildren();
    int n = children.length;
    // update display
    setCurrent(null);
    for (int i=0; i<n; i++)
    {
      if (!children[i].isClosed())
      {
        setCurrent(children[i]);
        break;
      }
    }
    return true;
  }
  
  /***************************************************************************
   * Inform proof that state was extended by children.
   * @param state a proof state that was extended by children
   * @param children open children states of state 
   **************************************************************************/
  public void extend(ProofState state, ProofState[] children)
  {
    int pos = open.indexOf(state);
    open.remove(state);
    int n = children.length;
    int p = 0;
    ProofState[] replayChildren = State.getReplayChildren();
    for (int i=0; i<n; i++)
    {
      ProofState cstate = children[i];
      states.put(cstate.getLabel(), cstate);
      // children may need inconsistency check during proof replay
      if (replayChildren != null && n == replayChildren.length) 
        State.setReplayChildren(replayChildren[i].getChildren());
      boolean oldAuto = State.useAutoSimplify();
      State.setAutoSimplify(cstate.useAutoSimplify());
      cstate.simplify(false);
      State.setAutoSimplify(oldAuto);
      State.setReplayChildren(replayChildren);
      if (cstate.isClosed())
        out.println("Proof state [" + cstate.getLabel() + 
        "] is closed by decision procedure.");
      else
      {
        open.add(pos+p, cstate);
        p = p+1;
      }
    }
    if (open.isEmpty())
    {
      closed = true;
      setCurrent(null);
      return;
    }
    if (pos == open.size()) pos = 0;
    setCurrent((ProofState)open.elementAt(pos));
  }
  
  /***************************************************************************
   * switch current state to next open state in sequence.
   **************************************************************************/
  public void nextOpenState()
  {
    int pos = open.indexOf(current)+1;
    if (pos == open.size())
      pos = 0;
    setCurrent((ProofState)open.elementAt(pos));
  }
  
  /***************************************************************************
   * switch current state to previous open state in sequence.
   **************************************************************************/
  public void prevOpenState()
  {
    int pos = open.indexOf(current)-1;
    if (pos == -1)
      pos = open.size()-1;
    setCurrent((ProofState)open.elementAt(pos));
  }
  
  /***************************************************************************
   * Print status of proof to out
   * @param out output stream.
   **************************************************************************/
  public void printStatus(PrintWriter out)
  {
    out.print("proof status: ");
    if (trusted)
      out.print("trusted");
    else
    {
      out.print("untrusted");
      return;
    }
    if (closed)
      out.print(", closed");
    else
    {
      out.print(", open");
      return;
    }
    if (absolute)
      out.print(", absolute");
    else
      out.print(", relative");
  }
  
  /***************************************************************************
   * Get status of proof.
   **************************************************************************/
  public String getStatus()
  {
    StringWriter s = new StringWriter();
    printStatus(new PrintWriter(s));
    return s.toString();
  }
  
  /***************************************************************************
   * Print proof.
   **************************************************************************/
  public void print()
  {
    printStatus(out);
    out.println();
    root.print(out, 0);
  }
  
  /***************************************************************************
   * 
   * External proof representation.
   * 
   **************************************************************************/
  
  // currently used document in DOM node generation
  private Document document;
  
  // global symbols referenced in DOM node generation
  private HashSet referenced;
  
  /***************************************************************************
   * Get collection of symbols of those entities referenced
   * in the last call of toNode().
   * @return collection of Symbol objects.
   **************************************************************************/
  public Collection getReferenced()
  {
    return referenced;
  }
  
  /****************************************************************************
   * Mark identifier as referenced in current proof.
   * @param ident the identifier that is marked as referenced.
   **************************************************************************/
  public void referenced(Identifier ident)
  {
    if (!(ident instanceof Reference)) return;
    Symbol isymbol = ((Reference)ident).getSymbol();
    if (isymbol instanceof TypeSymbol)
      referenced.add(isymbol);
    else if (isymbol instanceof FormulaSymbol)
      referenced.add(isymbol);
    else if (isymbol instanceof ValueSymbol)
    {
      if (symbol.getEnvironment().getValueSymbols().contains(isymbol)) 
        referenced.add(isymbol);
    }
    else
      State.setError("unknown symbol: " + ident.toString());
  }
  
  /***************************************************************************
   * Convert proof to DOM node in denoted document.
   * @param document the document to which the proof belongs.
   * @return the DOM node of the proof in this document.
   **************************************************************************/
  public Node toNode(Document document)
  {
    this.document = document;
    this.referenced = new HashSet();
    this.referenced.add(symbol);
    Element node = document.createElementNS(
      State.NamespaceURI, State.NamespacePrefix + ":proof");
    node.setAttribute("name", symbol.getIdentifier().getName());
    node.setAttribute("closed", Boolean.toString(closed));
    node.setAttribute("autosimplify", 
      Boolean.toString(getRoot().useAutoSimplify()));
    node.setAttribute("xmlns:" + State.NamespacePrefix, State.NamespaceURI);
    node.setAttribute("xmlns:" + State.OpenMathPrefix, State.OpenMathURI);
    addStates(node, root);
    return node;
  }
  
  /****************************************************************************
   * Add to node the DOM representation of state and its children.
   * @param node the DOM node to append the representation to.
   * @param state the state which is to be represented.
   ***************************************************************************/
  private void addStates(Node node, ProofState state)
  {
    Node node0 = document.createElementNS(
      State.NamespaceURI, State.NamespacePrefix + ":state");
    node.appendChild(node0);
    Command command = state.getCommand();
    if (command == null) 
    { 
      node0.appendChild(new Empty().toNode(document));
      return;
    }
    node0.appendChild(command.toNode(document));
    ProofState[] children = state.getChildren();
    if (children == null) return;
    int n = children.length;
    for (int i=0; i<n; i++)
      addStates(node0, children[i]);
  }
  
  /***************************************************************************
   * Construct skeleton proof from DOM representation.
   * @param symbol the symbol of the formula which is proved.
   * @param node a DOM node representing the proof.
   * @return the proof (null, if an error occured).
   *************************************************************************/
  public static Proof toProof(FormulaSymbol symbol, Node node)
  {
    String name = node.getLocalName();
    if (!name.equals("proof"))
    {
      State.setError("node has name " + name + " (\"proof\" expected)");
      return null;
    }
    if (!(node instanceof Element))
    {
      State.setError("element expected: " + node.toString());
      return null;
    }
    Element element = (Element)node;
    String cstring = element.getAttribute("closed");
    if (cstring == "")
    {
      State.setError("attribute 'closed' expected: " + element.toString());
      return null;
    }
    boolean closed;
    if (cstring.equals("true"))
      closed = true;
    else if (cstring.equals("false"))
      closed = false;
    else
    {
      State.setError("attribute 'closed' must be boolean: " + cstring);
      return null;
    }
    String astring = element.getAttribute("autosimplify");
    boolean autoSimplify;
    if (astring == "" || astring.equals("true"))
      autoSimplify = true;
    else if (astring.equals("false"))
      autoSimplify = false;
    else
    {
      State.setError("attribute 'autosimplify' must be boolean: " + astring);
      return null;
    }
    NodeList children = element.getChildNodes();
    int n = children.getLength();
    if (n != 1)
    {
      State.setError("one child expected: " + element.toString());
      return null;
    }
    Proof proof = new Proof();
    proof.setClosed(closed);
    proof.symbol = symbol;
    symbol.setProof(proof);
    proof.root = 
      new ProofState(proof, symbol.getEnvironment(), symbol.getFormula(),
          autoSimplify);
    boolean okay = toState(children.item(0), proof.root);
    if (!okay) return null;
    return proof;
  }
  
  /***************************************************************************
   * Construct proof state tree from node.
   * @param node a DOM node representing the proof state.
   * @param state the state to be filled with the information of the node
   * @return true iff no problem occured in construction.
   **************************************************************************/
  private static boolean toState(Node node, ProofState state)
  {
    String name = node.getLocalName();
    if (!name.equals("state"))
    {
      State.setError("node has name " + name + " (\"state\" expected)");
      return false;
    }
    if (!(node instanceof Element))
    {
      State.setError("element expected: " + node.toString());
      return false;
    }
    Element element = (Element)node;
    NodeList children = element.getChildNodes();
    int n = children.getLength();
    if (n == 0)
    {
      State.setError("children expected: " + node.toString());
      return false;
    }
    Node child = children.item(0);
    if (!(child instanceof Element))
    {
      State.setError("element expected: " + child.toString());
      return false;
    }
    
    Command command = CommandBase.toCommand((Element)child);
    if (command == null) return false;
    if (command instanceof Empty) command = null;
    
    Proof proof = state.getProof();
    ProofState[] cstate = new ProofState[n-1];
    for (int i=0; i<n-1; i++)
      cstate[i] = new ProofState(proof);
    state.setChildren(command, cstate);
    state.labelChildren();
    
    for (int i=0; i<n-1; i++)
    {
      boolean okay = toState(children.item(i+1), cstate[i]);
      if (!okay) return false;
    }
    return true;
  }
  
  // flag that signals whether proof replay had intermediate error
  private boolean replayError;
  
  /***************************************************************************
   * Replay skeleton proof.
   * @return true iff no error occured in replay
   **************************************************************************/
  public boolean replay()
  {
    if (!skeleton) return true;
    boolean oldTrusted = isTrusted();
    
    Command command = root.getCommand();
    ProofState[] children = root.getChildren();
    
    // existence of command signals that state is closed
    root.setChildren(null, null); 
    boolean okay = initialize();
    if (!okay) return false;
    okay = isClosed();
    
    // associate proof to tree display when it has no children
    // (otherwise whole tree is shown which can take very long)
    Proof proof = State.getProof();
    State.setProof(this);
    root.setChildren(command, children);
    
    // remember whether error occured
    replayError = false;
    
    if (!okay) okay = replay(root);
    State.setProof(proof);
      
    if (okay)
    {
      if (replayError && isClosed())
      {
        // proof replay had error that was fixed by user
        out.println("Fixed proof is written to store.");
        Store store = State.getStore();
        okay = store.write(this);
      }
      else if (!oldTrusted && isClosed())
      {
        // proof replay had error that was fixed by user
        out.println("Proof is trusted again.");
        Store store = State.getStore();
        okay = store.write(this);
      }
    }
    
    // generate index file
    File file = State.getPresenter().writeProofIndex(this);
    if (file == null) return false;
    
    return okay;
  }
  
  /***************************************************************************
   * Replay skeleton proof on proof state and its children.
   * @param state the state on which the proof is to be replayed.
   * @return true iff no error occured in replay.
   **************************************************************************/
  public boolean replay(ProofState state)
  {
    Command command = state.getCommand();
    if (command == null) return true;
    
    ProofState[] children = state.getChildren();
    int n = 0;
    if (children != null) n = children.length;
    
    state.setChildren(null, null);
    setCurrent(state);
    while (true)
    { 
      // show currently attempted proof command
      //    "Proved" must be set by simplifier
      if (!(command instanceof Proved))
      {
        state.setCommand(command);
        state.hasChanged(false);
      }
      
      // speed up proof replay by not attempting to close states with children.
      State.setReplayChildren(children);
      boolean auto = State.useAutoSimplify();
      State.setAutoSimplify(state.useAutoSimplify());
      command.process();
      State.setAutoSimplify(auto);
      State.setReplayChildren(null);
      if (State.getError() == null) break;

      // error has occurred
      replayError(state, State.getError());
      state.print(out);
      out.println("Command \"" + command.toString() + "\" causes an error.");
      State.resetError();
      command = readCommand();
      if (command == null) return false;
    }
    
    ProofState[]children0 = state.getChildren();
    int n0 = 0;
    if (children0 != null) n0 = children0.length;
    if (n0 == n)
    {
      boolean okay = true;
      for (int i=0; i<n; i++)
      {
        ProofState child = children[i];
        ProofState child0 = children0[i];
        if (child0.isClosed()) continue;
        child0.setChildren(child.getCommand(), child.getChildren());
        boolean okay0 = replay(child0);
        if (!okay0) okay = false;
      }
      return okay;
    }
    
    // check for failure of automatic instantiation
    if (command instanceof Auto && n0 == 0)
    {
      replayError(state, "command \"auto\" failed"); 
      return false;
    }
    
    out.println("Warning: problem in state [" + state.getLabel() + 
      "]."); 
    out.println("In the old proof, command \"" + command.toString() + 
      "\" generated " + n + " states:");
    for (int i=0; i<n; i++)
    {
      Command command0 = children[i].getCommand();
      String cstring = command0 == null ? "(empty)" : command0.toString();
      out.println(i + ":\t" + cstring);
    }
    
    if (n0 == 0)
    {
      out.print("In the new proof, the command generates no states"); 
      return false;
    }
    
    out.print("In the new proof, the command generates " + n0 + 
        " states with the following goals:");
    for (int i=0; i<n0; i++)
      out.println(i + ":\t" + getGoalString(children0[i]));
      
    out.println("You can now map ranges [a..b] of new states " +
                "to equally sized ranges [c..] of old states.");       
    int[] mapping = readMapping(n0, n);
    if (mapping == null) return false;
    
    boolean okay = true;
    for (int i=0; i<n0; i++)
    {
      int m = mapping[i];
      if (m == -1) continue;
      ProofState child = children[m];
      ProofState child0 = children0[i];
      if (child0.isClosed()) continue;
      child0.setChildren(child.getCommand(), child.getChildren());
      boolean okay0 = replay(child0);
      if (!okay0) okay = false;
    }
    return okay;
  }
  
  /****************************************************************************
   * Report error in replay of denoted state.
   * @param state state whose replay was attempted.
   * @param error a description of the error.
   ***************************************************************************/
  private void replayError(ProofState state, String error)
  {
    state.setCommand(null);
    state.setChildren(null, null);
    setCurrent(state);
    open.add(state);
    closed = false;
    replayError = true;
    out.println("Error in state ["  + state.getLabel() + "]: " + error + "."); 
  }
  
  /***************************************************************************
   * Ask user for a substitution command.
   * @return command or null (if no command was given)
   **************************************************************************/
  private static Command readCommand()
  {
    while (true)
    {
      boolean yes = Main.ask("Do you want to substitute the command");
      if (!yes) return null;
      out.println("prove> ");
      Command command = Main.parse();
      if (command != null) return command;
      out.println("Syntax error: " + State.getError());
      State.resetError();
    }
  }  
  
  /***************************************************************************
   * Read mapping from n old states to n0 new states.
   * @param n0 number greater equal 0.
   * @param n number greater equal 0.
   * @return mapping [0..n0-1] to [0..n-1] union {-1}
   *         such that no two different arguments are mapped to the
   *         same result different from -1.
   *         the result may also be null, which signals an error.
   **************************************************************************/
  private static int[] readMapping(int n0, int n)
  {
    while (true)
    {
      // mapping maps new indices to old ones, used is the inverse mapping
      int[] mapping = new int[n0];
      for (int i=0; i<n0; i++) mapping[i] = -1;
      int[] used = new int[n];      
      for (int i=0; i<n; i++) used[i] = -1;
      boolean mapped = false;
      while (true)
      {
        int a = readNumber("Enter start state a", 0, n0-1);
        if (a == -1) break;
        int s = Math.min(n0-a, n);
        int b = readNumber("Enter end state b", a, a+s-1);
        if (b == -1) break;
        s = b-a+1;
        int c = readNumber("Enter start state c", 0, n-s);
        if (c == -1) break;
        boolean found = false;
        for (int i=0; i<s; i++)
        {
          found = used[c+i] != -1;
          if (found)
          {
            out.println("Old state " + (c+i) + " cannot be used because " +
              " the new state " + used[c+i] + " is already mapped to it.");
            break;
          }
        }
        if (found) continue;
        mapped = true;
        for (int i=0; i<s; i++)
        {
          mapping[a+i] = c+i;
          used[c+i] = a+i;
        }
      }
      if (!mapped)
      {
        out.println("You have not constructed any mapping.");
        boolean yes = Main.ask("Do you want to continue without mapping");
        if (yes) return null;
        continue;
      }
      out.println("You have constructed the following mapping:");
      int a = -1;
      for (int i=0; i<n0; i++)
      {
        if (a == -1)
        {
          if (mapping[i] != -1) a = i;
        }
        else if (mapping[i] == -1 || (i-a != mapping[i]-mapping[a]))
        {
          if (a == i-1)
            out.println("New state " + a + " is mapped " +
                "to old state " + mapping[a] + ".");
          else
            out.println("New states [" + a + ".." + (i-1) + "] are mapped " +
              "to old states [" + mapping[a] + ".." + mapping[i-1] + "].");
          if (mapping[i] == -1) 
            a = -1;
          else
            a = i;
        }
      }
      if (a != -1)
      {
        if (a == n0-1)
          out.println("New state " + a + " is mapped " +
              "to old state " + mapping[a] + ".");
        else
          out.println("New states [" + a + ".." + (n0-1) + "] are mapped " +
              "to old states [" + mapping[a] + ".." + mapping[n0-1] + "].");
      }
      boolean yes = Main.ask("Do you want to use this mapping");
      if (yes) return mapping;
    }
  }
  
  /***************************************************************************
   * Ask user for a number in the range a..b.
   * @param prompt the message printed for asking. 
   * @param a the lower bound of the number.
   * @param b the upper bound of the number.
   * @return number in range a..b or -1 (if no number was given)
   **************************************************************************/
  private static int readNumber(String prompt, int a, int b)
  {
    while (true)
    {
      out.print(prompt + " (" + a + ".." + b + ", -1 to abort): ");
      String line = Main.readLine();
      if (line == null) return -1;
      try
      {
        int pos = Integer.parseInt(line);
        if (pos == -1 || (a <= pos && pos <= b)) return pos;
      }
      catch (NumberFormatException e)
      { }
    }
  } 
 
 
  /***************************************************************************
   * Return representation of last goal of state.
   * @param state a proof state.
   * @return a string representation of the last goal.
   **************************************************************************/
  private static String getGoalString(ProofState state)
  {
    Formula[] goals = state.getGoals();
    int l = goals.length;
    if (l == 0) return "(none)";
    return goals[l-1].getExpression().toString();
  }
}
